##
##  pullDemo.py: sample wrappers for pull API
##  Copyright (C) 2010, 2009, Gordon Kindlmann
##
##  Permission is hereby granted, free of charge, to any person obtaining
##  a copy of this software and associated documentation files (the
##  "Software"), to deal in the Software without restriction, including
##  without limitation the rights to use, copy, modify, merge, publish,
##  distribute, sublicense, and/or sell copies of the Software, and to
##  permit persons to whom the Software is furnished to do so, subject to
##  the following conditions:
##
##  The above copyright notice and this permission notice shall be
##  included in all copies or substantial portions of the Software.
##
##  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
##  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
##  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
##  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
##  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
##  OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
##  WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
##

'''
This is really hastily written, and doesn't conform in the least
to good conventions of Python coding. The worst part is using exit()
instead of raising exceptions when tings go wrong.

The "API" here (this python interface to the pull library) is not well
thought-out, and won't be helpful for integrating particles in other
applications.  Better designed interfaces should come in the future,
as more and more of Teem develops idiomatic python interfaces.

The code here is mainly to serve as a demo of using the pull library
in Teem (whose API should be quite stable), and to simplify
documenting using pull to produce typical results.
'''

import teem
import ctypes
import sys

##
## volLoad: loads in volumes
##
def volLoad(vlist, cachePath, kssBlurStr, verbose):
    kSSblur = teem.nrrdKernelSpecNew()
    if (teem.nrrdKernelSpecParse(kSSblur, kssBlurStr)):
        estr = teem.biffGetDone('nrrd')
        print "problem with kernel:\n%s" % estr
        sys.exit(1)
    vol = (ctypes.POINTER(teem.meetPullVol) * len(vlist))()
    E = 0
    for i in range(len(vlist)):
        vol[i] = teem.meetPullVolNew()
        if (not E):
            E += teem.meetPullVolParse(vol[i], vlist[i])
    if (not E):
        E += teem.meetPullVolLoadMulti(vol, len(vol),
                                       cachePath, kSSblur,
                                       teem.nrrdBoundaryWrap, 0.0,
                                       verbose)
    if (E):
        estr = teem.biffGetDone('meet')
        print "problem with volumes:\n%s" % estr
        sys.exit(1)
    teem.nrrdKernelSpecNix(kSSblur)
    return vol

def volFree(vol):
    for i in range(len(vol)):
        vol[i] = teem.meetPullVolNix(vol[i])

##
## especParse: returns a length-4 list of the arguments to pass
## to pullInterEnergySet. Arguments are:
## type: from pullIterType* enum
## r: string definition for energy function along space (especR)
## s: string definition for energy function along scale (especS)
## w: string definition for windowing energy function (especW)
##
def energyParse(**kwargs):
    type = kwargs['type']
    if (teem.pullInterTypeJustR == type
        or teem.pullInterTypeUnivariate == type):
        espec = [teem.pullEnergySpecNew(), None, None]
    elif (teem.pullInterTypeSeparable == type):
        espec = [teem.pullEnergySpecNew(),
                 teem.pullEnergySpecNew(), None]
    elif (teem.pullInterTypeAdditive == type):
        espec = [teem.pullEnergySpecNew(),
                 teem.pullEnergySpecNew(),
                 teem.pullEnergySpecNew()]
    lett = ['r', 's', 'w']
    E = 0
    for i in range(3):
        if (not E and espec[i]):
            E += teem.pullEnergySpecParse(espec[i], kwargs[lett[i]])
    if (E):
        estr = teem.biffGetDone('pull')
        print "problem with infos:\n%s" % estr
        sys.exit(1)
    ret = [type, espec[0], espec[1], espec[2]]
    return ret

def energyFree(espec):
    espec[1] = teem.pullEnergySpecNix(espec[1])
    espec[2] = teem.pullEnergySpecNix(espec[2])
    espec[3] = teem.pullEnergySpecNix(espec[3])
    return

##
## initSet: calls (and returns value from) the appropriate
## pullInitMethod* function. Possibilities are:
## [teem.pullInitMethodRandom, num]
## [teem.pullInitMethodPointPerVoxel,
##     ppv, zslcmin, zslcmax, alongscalenum, jitter]
## [teem.pullInitMethodGivenPos, npos]
##
def initSet(pctx, info):
    if (teem.pullInitMethodRandom == info[0]):
        return teem.pullInitRandomSet(pctx, info[1])
    elif (teem.pullInitMethodPointPerVoxel == info[0]):
        return teem.pullInitPointPerVoxelSet(pctx, info[1],
                                             info[2], info[3],
                                             info[4], info[5])
    elif (teem.pullInitMethodGivenPos == info[0]):
        return teem.pullInitGivenPosSet(pctx, info[1])


## place for storing pullContext->sysParm.gamma
gamma = 0

## way of getting arguments that allows for checking whether
## there were any bogus options
def a(args, a, default=None):
    if default is None:
        ret = args.get(a)
    else:
        ret = args.get(a, default)
    if a in args:
        del args[a]
    return ret

##
## All the set-up and commands that are required to do a pullContext run
##
def run(nposOut, **args):
    global gamma
    if (not ('vol' in args and 'info' in args and 'efs' in args)):
        print "run: didn't get vol, info, and efs args"
        sys.exit(1)

    # learn args, with defaults
    vol = a(args, 'vol')
    infoStr = a(args, 'info')
    efs = a(args, 'efs')
    init = a(args, 'init', [teem.pullInitMethodRandom, 100])
    energyDict = a(args, 'energy', {'type':teem.pullInterTypeJustR,
                                    'r':'qwell:0.68'})
    verbose = a(args, 'verbose', 1)
    rngSeed = a(args, 'rngSeed', 42)
    nobin = a(args, 'nobin', False)
    noadd = a(args, 'noadd', False)
    ac3c = a(args, 'ac3c', False)
    usa = a(args, 'usa', False)
    nave = a(args, 'nave', True)
    cbst = a(args, 'cbst', True)
    ratb = a(args, 'ratb', True)
    lti = a(args, 'lti', False)
    npcwza = a(args, 'npcwza', False)
    ubfgl = a(args, 'ubfgl', False)
    iterMin = a(args, 'iterMin', 0)
    iterMax = a(args, 'iterMax', 100)
    snap = a(args, 'snap', 0)
    pcp = a(args, 'pcp', 5)
    iad = a(args, 'iad', 10)
    radSpace = a(args, 'radSpace', 1)
    radScale = a(args, 'radScale', 1)
    alpha = a(args, 'alpha', 0.5)
    beta = a(args, 'beta', 0.5)
    gamma = a(args, 'gamma', 1)
    stepInitial = a(args, 'step', 1)
    ssBack = a(args, 'ssBack', 0.2)
    ssOppor = a(args, 'ssOppor', 1.1)
    pbm = a(args, 'pbm', 50)
    k00Str = a(args, 'k00', 'c4h')
    k11Str = a(args, 'k11', 'c4hd')
    k22Str = a(args, 'k22', 'c4hdd')
    kSSreconStr = a(args, 'kSSrecon', 'hermite')
    eip = a(args, 'eip', 0.00005)
    eiphl = a(args, 'eiphl', 0)
    fnnm = a(args, 'fnnm', 0.25)
    edmin = a(args, 'edmin', 0.00001)
    edpcmin = a(args, 'edpcmin', 0.01)
    maxci = a(args, 'maxci', 15)
    if args:
        print
        print "sorry, got unrecognized arguments:"
        print args
        sys.exit(1)

    # create pullContext and set up all its state.  Its kind of silly that
    # the pullContext is created anew everytime we want to run it, but thats
    # because currently the pullContext doesn't have the kind of internal
    # network of flags (like the gageContext does) that allow the context
    # to be modifyied and re-used indefinitely.  This will be fixed after
    # the Teem v1.11 release.
    pctx = teem.pullContextNew()
    energyList = energyParse(**energyDict)
    if (teem.pullVerboseSet(pctx, verbose) or
        teem.pullRngSeedSet(pctx, rngSeed) or
        teem.pullFlagSet(pctx, teem.pullFlagNixAtVolumeEdgeSpace, nave) or
        teem.pullFlagSet(pctx, teem.pullFlagConstraintBeforeSeedThresh,
                         cbst) or
        teem.pullFlagSet(pctx, teem.pullFlagEnergyFromStrength, efs) or
        teem.pullFlagSet(pctx, teem.pullFlagRestrictiveAddToBins, ratb) or
        teem.pullFlagSet(pctx, teem.pullFlagNoPopCntlWithZeroAlpha, npcwza) or
        teem.pullFlagSet(pctx, teem.pullFlagUseBetaForGammaLearn, ubfgl) or
        teem.pullFlagSet(pctx, teem.pullFlagBinSingle, nobin) or
        teem.pullFlagSet(pctx, teem.pullFlagNoAdd, noadd) or
        teem.pullFlagSet(pctx,
                         teem.pullFlagAllowCodimension3Constraints,
                         ac3c) or
        teem.pullInitUnequalShapesAllowSet(pctx, usa) or
        teem.pullIterParmSet(pctx, teem.pullIterParmMin, iterMin) or
        teem.pullIterParmSet(pctx, teem.pullIterParmMax, iterMax) or
        teem.pullIterParmSet(pctx, teem.pullIterParmSnap, snap) or
        teem.pullIterParmSet(pctx, teem.pullIterParmConstraintMax, maxci) or
        teem.pullIterParmSet(pctx, teem.pullIterParmPopCntlPeriod, pcp) or
        teem.pullIterParmSet(pctx, teem.pullIterParmAddDescent, iad) or
        teem.pullIterParmSet(pctx,
                             teem.pullIterParmEnergyIncreasePermitHalfLife,
                             eiphl) or
        teem.pullSysParmSet(pctx, teem.pullSysParmStepInitial, stepInitial) or
        teem.pullSysParmSet(pctx, teem.pullSysParmRadiusSpace, radSpace) or
        teem.pullSysParmSet(pctx, teem.pullSysParmRadiusScale, radScale) or
        teem.pullSysParmSet(pctx, teem.pullSysParmAlpha, alpha) or
        teem.pullSysParmSet(pctx, teem.pullSysParmBeta, beta) or
        teem.pullSysParmSet(pctx, teem.pullSysParmGamma, gamma) or
        teem.pullSysParmSet(pctx, teem.pullSysParmEnergyIncreasePermit, eip) or
        teem.pullSysParmSet(pctx, teem.pullSysParmFracNeighNixedMax, fnnm) or
        teem.pullSysParmSet(pctx, teem.pullSysParmEnergyDecreaseMin, edmin) or
        teem.pullSysParmSet(pctx, teem.pullSysParmEnergyDecreasePopCntlMin,
                            edpcmin) or
        teem.pullSysParmSet(pctx, teem.pullSysParmBackStepScale, ssBack) or
        teem.pullSysParmSet(pctx, teem.pullSysParmOpporStepScale, ssOppor) or
        teem.pullProgressBinModSet(pctx, pbm) or
        teem.pullInterEnergySet(pctx, energyList[0], energyList[1],
                                energyList[2], energyList[3])):
        estr = teem.biffGetDone('pull')
        print "problem with set-up:\n%s" % estr
        sys.exit(1)

    # The meet library offers a slightly higher-level abstraction of
    # the pullInfo; this code parses a bunch of info specifications
    info = (ctypes.POINTER(teem.meetPullInfo) * len(infoStr))()
    for i in range(len(infoStr)):
        info[i] = teem.meetPullInfoNew()
        if (teem.meetPullInfoParse(info[i], infoStr[i])):
            estr = teem.biffGetDone('meet')
            print "problem with infos:\n%s" % estr
            sys.exit(1)

    # Create all the kernels needed for reconstruction
    k00 = teem.nrrdKernelSpecNew()
    k11 = teem.nrrdKernelSpecNew()
    k22 = teem.nrrdKernelSpecNew()
    kSSrecon = teem.nrrdKernelSpecNew()
    if (teem.nrrdKernelSpecParse(k00, k00Str) or
        teem.nrrdKernelSpecParse(k11, k11Str) or
        teem.nrrdKernelSpecParse(k22, k22Str) or
        teem.nrrdKernelSpecParse(kSSrecon, kSSreconStr)):
        estr = teem.biffGetDone('nrrd')
        print "problem with kernels:\n%s" % estr
        sys.exit(1)

    # We do assume that the volumes came in loaded (as opposed to reloading
    # them here for every one); here we add all the volumes and infos into
    # the pullContext
    if (teem.meetPullVolAddMulti(pctx, vol, len(vol),
                                 k00, k11, k22, kSSrecon) or
        teem.meetPullInfoAddMulti(pctx, info, len(info))):
            estr = teem.biffGetDone('meet')
            print "problem with infos:\n%s" % estr
            sys.exit(1)

    # Final setup and the actual "pullRun" call
    if (initSet(pctx, init) or
        teem.pullInitLiveThreshUseSet(pctx, lti) or
        teem.pullStart(pctx) or
        teem.pullRun(pctx) or
        teem.pullOutputGet(nposOut, None, None, None, 0, pctx)):
        estr = teem.biffGetDone('pull')
        print "problem running system:\n%s" % estr
        sys.exit(1)

    # until there's a clean way to re-use a pullContext for different runs,
    # there's no good way for the user to call pullGammaLearn (because pctx
    # doesn't survive outside the call to "run".  So we call pullGammaLearn
    # it whenever it makes sense to do so and save the result in the global
    # "gamma"
    if (teem.pullInterTypeAdditive == pctx.contents.interType and
        teem.pullEnergyButterworthParabola.contents.name
        == pctx.contents.energySpecS.contents.energy.contents.name):
        if (teem.pullGammaLearn(pctx)):
            estr = teem.biffGetDone('pull')
            print "problem learning gamma:\n%s" % estr
            sys.exit(1)
        gamma = pctx.contents.sysParm.gamma
        print "pullGammaLearn returned", gamma

    teem.nrrdKernelSpecNix(k00)
    teem.nrrdKernelSpecNix(k11)
    teem.nrrdKernelSpecNix(k22)
    teem.nrrdKernelSpecNix(kSSrecon)
    energyFree(energyList)
    teem.pullContextNix(pctx)
    return



#! /usr/bin/env python

##
##  gen-teem.py: automatically-generated ctypes python wrappers for Teem
##  Copyright (C) 2013, 2012, 2011 University of Chicago
##  (originally created by Sam Quinan - samquinan@cs.uchicago.edu)
##
##  Permission is hereby granted, free of charge, to any person obtaining
##  a copy of this software and associated documentation files (the
##  "Software"), to deal in the Software without restriction, including
##  without limitation the rights to use, copy, modify, merge, publish,
##  distribute, sublicense, and/or sell copies of the Software, and to
##  permit persons to whom the Software is furnished to do so, subject to
##  the following conditions:
##
##  The above copyright notice and this permission notice shall be
##  included in all copies or substantial portions of the Software.
##
##  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
##  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
##  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
##  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
##  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
##  OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
##  WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
##

##
## for posterity; this records the contents of the sourceme-ctypeslib.txt
## file that this script automates ...
oldtodo="""
echo === These are commands that GLK used to create teem.py, in tcsh
echo === Obviously, anyone is welcome to make this into a proper python script
echo === or otherwise make it more robust and portable, especially the stuff
echo === that removes the install path specifics from teem.py
echo ===
echo === You will need a recent gccxml to make this work.

svn co http://svn.python.org/projects/ctypes/branches/ctypeslib-gccxml-0.9
setenv CTYPES `pwd`/ctypeslib-gccxml-0.9
setenv PYTHONPATH ${CTYPES}:`pwd`

echo === A patch is still needed, hopefully this can be fixed
patch -p0 -i ctypes-codegen.patch

echo === TEEM has to be set to wherever CMake put its make install
setenv TEEM /Users/gk/teem-install

echo === all.h just includes all the other public top-level .h in Teem
rm -rf teemincl
mkdir teemincl
cp -r ${TEEM}/include/teem teemincl
cd teemincl
ls -1 teem/*.h | awk '{print "#include <"$1">"}' > all.h
cd ..

echo === teem.xml represents the Teem API
python ${CTYPES}/scripts/h2xml.py `pwd`/teemincl/all.h -I teemincl -o teem.xml

echo === DYLD_LIBRARY_PATH is obviously Mac specific, please fix
setenv DYLD_LIBRARY_PATH ${TEEM}/lib
python ${CTYPES}/scripts/xml2py.py teem.xml -llibteem.dylib -o pre-teem.py -m stdio \
        -r "(air|hest|biff|nrrd|ell|unrrdu|alan|moss|tijk|gage|dye|bane|limn|echo|hoover|seek|ten|elf|pull|coil|push|mite|meet).*"
rm -rf teemincl


echo === At this point GLK does emacs stuff to make pre-teem.py into what is
echo === the currently svn committed teem.py
echo === Obviously, this really needs to be automated...
"""
# and now it has been automated!

import os, sys, shutil, platform, re, string

if len(sys.argv) != 3:
    sys.exit("usage: gen-teem.py <ctypeslib-gccxml source dir> <teem install dir>")

## (TEEM_LIB_LIST)
libs_list = ["air", "hest", "biff", "nrrd", "ell", "unrrdu", "alan", "moss", "tijk", "gage", "dye", "bane", "limn", "echo", "hoover", "seek", "ten", "elf", "pull", "coil", "push", "mite", "meet"]

#
# validate os
#
if os.name != "posix":
    sys.exit("program only supports posix compilant systems at this time")

#
# validate gccxml install
#
tmp = os.path.join(os.getcwd(), "tmp")
os.system("gccxml --version &> tmp")
file = open(tmp, "r")
file.seek(0)
first_line = file.readline()
file.close()
if not first_line.startswith("GCC-XML version"):
    os.remove(tmp)
    sys.exit("error: gccxml of version >= 0.9.0 is required for this script")

if not ((int(first_line[16]) > 0) or ((int(first_line[18]) == 9) and (int(first_line[20]) >= 0))):
    os.remove(tmp)
    sys.exit("error: gccxml of version >= 0.9.0 is required for this script")
os.remove(tmp)

#
# validate ctypeslib-gccxml source dir path
#
if not os.path.isdir(sys.argv[1]):
    sys.exit("%s does not point to a directory" % sys.argv[1])

CTYPES = os.path.abspath(sys.argv[1])
#
# NOTE, however, that this really assumes a PATCHED ctypeslib-gccxml,
# via a procedure like:
# svn co http://svn.python.org/projects/ctypes/branches/ctypeslib-gccxml-0.9
# patch -p0 -i ctypes-codegen.patch

#
# validate teem install dir path
#
if not os.path.isdir(sys.argv[2]):
    sys.exit("%s does not point to a directory" % sys.argv[2])

TEEM = os.path.abspath(sys.argv[2])


#
# copy files from install dir
#

TMP_DIR = "gen-teem-tmp"

teem_include = os.path.join(TEEM, "include")

if not os.path.isdir(teem_include):
    sys.exit("%s is not the teem install directory; %s is not a valid path" % (teem, teem-install))

if os.path.isdir(TMP_DIR):
    shutil.rmtree(TMP_DIR)

shutil.copytree(teem_include, TMP_DIR)


#
# generate all.h
#

all_h = os.path.join(TMP_DIR, "all.h")
f_open = open (all_h, "w")

defines = []

Files = os.listdir(os.path.join(TMP_DIR, "teem"))
for file in Files:
    if file.endswith(".h"):
        f_open.write("#include <teem/%s> %s" % (file, os.linesep))
        #
        # get #define statements from file
        #
        lines = open (os.path.join(TMP_DIR, "teem", file), "r").readlines()
        expr1 = re.compile("^#define")
        expr2 = re.compile("\\\\")
        expr3 = re.compile("HAS_BEEN_INCLUDED")
        expr4 = re.compile("##")
        expr5 = re.compile("^#define [^ ]*\(")
        expr6 = re.compile("NRRD_TYPE_BIGGEST")
        # this strips out NRRD_LLONG_MAX, NRRD_LLONG_MIN, and NRRD_ULLONG_MAX,
        # which depend on macros AIR_LLONG and AIR_LLONG.  For some reason
        # this causes a problem now (Thu Jun 14 11:49:14 CDT 2012) even though
        # they haven't before, though these constants and macros are not new
        expr7 = re.compile("LLONG")
        for line in lines:
            if (expr1.search(line) and not expr2.search(line)
                and not expr3.search(line) and not expr4.search(line)
                and not expr5.search(line) and not expr6.search(line)
                and not expr7.search(line)):
                firstword, restwords = string.replace(string.replace(line[8:], "/*", "#" ), "*/", "").split(None,1)
                defines.append("%s = %s" % (firstword, restwords))
f_open.close()

#
# generate teem.xml
#
# note: these calls are unix only because windows support for command line
#	append to python path (where the scopeis only that one call) is too
#	difficult at this time -- better support may be added to later versions
#	of python, so that something to watch for / modify in the future

pypath_append = "PYTHONPATH=%s:%s" % (CTYPES, os.getcwd())
teem_xml = os.path.join(os.getcwd(), "teem.xml")
os.system("%s python %s %s -I %s -o %s" % (pypath_append, os.path.join(CTYPES, "scripts", "h2xml.py"), os.path.abspath(all_h), TMP_DIR, teem_xml))


#
# generate pre-teem.py
#
pre_teem_py = os.path.join(os.getcwd(), "pre-teem.py")

teem_libs = '|'.join(libs_list)

system_type = platform.system()
dll_path = ""
ext = ""
substr = ""
if system_type == "Darwin":
    dll_path = "DYLD_LIBRARY_PATH=%s" % os.path.join(TEEM, "lib")
    ext = "dylib"
    substr = "_libraries['%s']" % os.path.join(TEEM, "lib", "libteem.dylib")
else:
    dll_path = "LD_LIBRARY_PATH=%s" % os.path.join(TEEM, "lib")
    ext = "so"
    substr = "_libraries['libteem.so']"

os.system("%s %s python %s %s -l libteem.%s -o %s -m stdio -r \"(%s).*\"" % (dll_path, pypath_append, os.path.join(CTYPES, "scripts", "xml2py.py"), teem_xml, ext,  pre_teem_py, teem_libs))

#
# generate teem.py
#

libs_destuctable = list(libs_list)

contents = open(pre_teem_py, "r").readlines()[8:]
mod_contents = []
for line in contents:
    l = line.replace(substr, "libteem")
    if not (("__darwin_size_t = c_ulong" in l) or ("size_t = __darwin_size_t" in l)):
        l2 = l.replace("\'__darwin_size_t\',", "") #designed to remove defs from long list at end
        l3 = l2.replace("\'size_t\',", "") #designed to remove defs from long list at end
        l = l3
        if not re.compile("_size_t").search(l3):
            if not re.match("#", l3): # i.e. do not make changes in commented lines
                l = l3.replace("size_t", "c_size_t")
        mod_contents.append(l)
        if "Present" in l:
            for lib in libs_destuctable:
                lib_str = lib+"Present"
                if lib_str in l:
                    libs_destuctable.remove(lib)
                    break

# in experimental libs not included, cleanup and fail
if libs_destuctable: # empty sequence implicity false
    shutil.rmtree(TMP_DIR)
    os.remove(teem_xml)
    #os.remove(pre_teem_py)
    sys.exit("ERROR: experimental libs: %s not turned on - please rebuild teem with BUILD_EXPERIMENTAL_LIBS turned on, then re-run gen-teem.py" % ','.join(libs_destuctable))

header = [
"##",
"##  teem.py: automatically-generated ctypes python wrappers for Teem",
"##  Copyright (C) 2013, 2012, 2011, 2010, 2009  University of Chicago",
"##",
"##  Permission is hereby granted, free of charge, to any person obtaining",
"##  a copy of this software and associated documentation files (the",
"##  \"Software\"), to deal in the Software without restriction, including",
"##  without limitation the rights to use, copy, modify, merge, publish,",
"##  distribute, sublicense, and/or sell copies of the Software, and to",
"##  permit persons to whom the Software is furnished to do so, subject to",
"##  the following conditions:",
"##",
"##  The above copyright notice and this permission notice shall be",
"##  included in all copies or substantial portions of the Software.",
"##",
"##  THE SOFTWARE IS PROVIDED \"AS IS\", WITHOUT WARRANTY OF ANY KIND,",
"##  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF",
"##  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND",
"##  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE",
"##  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION",
"##  OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION",
"##  WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.",
"##",
"##############################################################",
"##############################################################",
"#### NOTE: This teem.py file is automatically produced by",
"#### teem/python/ctypes/gen-teem.py.  Necessary changes to ",
"#### teem.py should be made in gen-teem.py, not here.",
"##############################################################",
"##############################################################",
"",
"from ctypes import *",
"import ctypes.util",
"import sys, os",
"",
"def load_library(libname, loader_path=\"\"):",
"    ext = os.path.splitext(libname)[1]",
"    if not ext:",
"        # Try to load library with platform-specific name",
"        if sys.platform == 'win32':",
"            libname_ext = '%s.dll' % libname",
"        elif sys.platform == 'darwin':",
"            libname_ext = '%s.dylib' % libname",
"        elif sys.platform == 'linux2':",
"            libname_ext = '%s.so' % libname",
"    else:"
"        libname_ext = libname",
"",
"    if (loader_path != \"\"):",
"        loader_path = os.path.abspath(loader_path)",
"        if not os.path.isdir(loader_path):",
"            libdir = os.path.dirname(loader_path)",
"        else:",
"            libdir = loader_path",
"    else:",
"        libdir = loader_path",
"",
"    try:",
"        libpath = os.path.join(libdir, libname_ext)",
"        return CDLL(libpath)",
"    except OSError, e:",
"        raise e"
"",
"try:",
"    libteem = load_library('libteem')",
"except OSError:",
"    print \"**\"",
"    print \"**  teem.py couldn't find and load the \\\"libteem\\\" shared library.\"",
"    print \"**\"",
"    print \"**  try setting optional loader_path argument in the load_library() call above to '<teem-install-dir>/lib/'\"",
"    print \"**\"",
"    raise ImportError",
"",
"# =============================================================",
"# Utility types and classes to help teem.py be platform-independent.",
"",
"STRING = c_char_p",
"",
"class FILE(Structure):",
"    pass",
"",
"# oddly, size_t is in ctypes, but not ptrdiff_t",
"# which is probably a bug",
"if sizeof(c_void_p) == 4:",
"    ptrdiff_t = c_int32",
"elif sizeof(c_void_p) == 8:",
"    ptrdiff_t = c_int64",
"",
"# =============================================================",
"# What follows are all the functions, struct definitions, globals, ",
"# enum values, and typedefs in Teem. This is generated by ctypeslib:",
"#   http://svn.python.org/projects/ctypes/branches/ctypeslib-gccxml-0.9",
"# followed by further post-processing and filtering.",
"# See end of this file for definitions of stderr, stdin, stdout",
]

footer = [
"# =============================================================",
"# Make sure this shared library will work on this machine.",
"if not nrrdSanity():",
"    errstr = biffGetDone(NRRD)",
"    print \"**\"",
"    print \"** Sorry, there is a problem (described below) with the \"",
"    print \"** Teem shared library that prevents its use. This will \"",
"    print \"** have to be fixed by recompiling the Teem library for \"",
"    print \"** this platform. \"",
"    print \"**\"",
"    print \"** %s\" % errstr",
"    raise ImportError",
"",
"# =============================================================",
"# Its nice to have these FILE*s around for utility use, but they ",
"# aren't available in a platform-independent way in ctypes. These ",
"# air functions were created for this purpose.",
"stderr = airStderr()",
"stdout = airStdout()",
"stdin = airStdin()",
]

teem_py = os.path.join(os.getcwd(), "teem.py")

if os.path.exists(teem_py):
    os.remove(teem_py)

out = open(teem_py, "w")

for line in header:
    out.write(line)
    out.write(os.linesep)

out.writelines(mod_contents)
out.write(os.linesep)

out.write("# =============================================================\n")
out.write("# What follows are the all #define's in Teem, excluding macros,\n")
out.write("# and #defines that depend on compile-time tests done by the\n")
out.write("# C pre-processor.\n")
out.write("# This is created by something akin to grep'ing through the\n")
out.write("# public header files, with some extra filters.\n")
out.write(os.linesep)
out.writelines(defines)
out.write(os.linesep)

for line in footer:
    out.write(line)
    out.write(os.linesep)

out.close()


#
# cleanup
#

shutil.rmtree(TMP_DIR)
os.remove(teem_xml)
os.remove(pre_teem_py)




import sys
import teem
import pullDemo
import ctypes

npos = teem.nrrdNew()
vol = pullDemo.volLoad(['../../data/fmob-c4h.nrrd:scalar:V'], './', 'ds:1,5', 1)
infoList = ['h-c:V:val:0:-1',
            'hgvec:V:gvec',
            'hhess:V:hess',
            'tan1:V:hevec2',
            'strn:V:heval2:0:-1',
            'lthr:V:heval2:-70:-1',
            'sthr:V:heval2:-70:-1']

pullDemo.run(npos,
             init=[teem.pullInitMethodRandom, 300],
             efs=False,
             vol=vol,
             info=infoList,
             energy={'type':teem.pullInterTypeJustR, 'r':'cwell:0.6,-0.002'},
             iterMax=200,
             radSpace=0.05)

pullDemo.volFree(vol)

if (teem.nrrdSave("npos.nrrd", npos, None)):
    estr = teem.biffGetDone('nrrd')
    print "problem running system:\n%s" % estr
    sys.exit(1)



#!/usr/bin/env python

# test.py: 16-bit grayscale PNG to nrrd file
# Sam Quinan

import Nrrd as nrd, numpy as np

# TESTING: fmob-ch4 crop test
print "TEST: fmob-ch4.nrrd"
a = nrd.Nrrd()
a.load("../../data/fmob-c4h.nrrd")
print "nrrd loaded..."
b = nrd.ExtendedArray(a)
print "wrapped as array..."
c = b[5:11,:,:]
print "cropped..."
#print c.shape
d = np.ascontiguousarray(c, dtype=c.dtype.type)
print "made contiguous..."
#print d.shape
e = nrd.Nrrd()
e.fromNDArray(d)
print "loaded..."
e.save("fmob_test_out.nrrd")
print "done"



# OLD TESTING -- IGNORE

#in_image = Image.open("msix.png")
#f_data = numpy.asarray(in_image)
#f_nrrd = Nrrd()
#f_nrrd.fromNDArray(f_data)
#del f_data
#f_nrrd.save("test.nrrd")
#t = f_nrrd
#a = numpy.asarray(f_nrrd)
#print a
#del f_nrrd

#a = Nrrd()
#a.load("msix.png")
#
#b = ExtendedArray(a)
#print b
#print type(b)
#print b.flags
#print b.data
#print b.base_ref
#print b.base is None
#
#print "-------------------------------"
#
#in_image = Image.open("msix.png")
#c = numpy.asarray(in_image)
#del in_image
#
#print c
#print type(c)
#print c.flags
#print c.data
#
#d = Nrrd()
#d.fromNDArray(c)
#
#e = ExtendedArray(d)
#e2 = e[:,[0,2]]
#
#print e
#print type(e)
#print e.flags
#print e.data
#print e.base_ref
#
#
#print "-------------------------------"
#
#arr = numpy.zeros((4,))
#print arr.base is None
#print arr
#v1 = arr.view(ExtendedArray)
#print v1.base is arr
#print v1
#v2 = arr[2:]
#print v2.base is arr
#print v2

#c_data = numpy.asarray(in_image)
#print "C CONTIGUOUS"
#print c_data
#print c_data.flags
#print c_data.strides
#f_data = numpy.require(in_image, dtype=None, requirements='F_CONTIGUOUS')
#print "F CONTIGUOUS"
#print f_data
#print f_data.flags
#print f_data.strides
#c_nrrd = ndarrayToNrrd(c_data)
#f_nrrd = ndarrayToNrrd(f_data)
#teem.nrrdSave("foo.nrrd", c_nrrd, None)
#teem.nrrdSave("bar.nrrd", f_nrrd, None)


from ctypes import *
from ctypes.util import find_library

class FILE(Structure):
    pass

# oddly, size_t is in ctypes, but not ptrdiff_t
if sizeof(c_void_p) == 4:
    ptrdiff_t = c_int32
elif sizeof(c_void_p) == 8:
    ptrdiff_t = c_int64

#libc = CDLL(find_library("c"))
# HEY these are probably Mac-specific, please fix
#stdin = POINTER(FILE).in_dll(libc, "__stdinp")
#stderr = POINTER(FILE).in_dll(libc, "__stderrp")
#stdout = POINTER(FILE).in_dll(libc, "__stdoutp")





#!/usr/bin/env python
##
##  Nrrd.py: bridge between Nrrd and Numpy arrays
##  Copyright (C) 2013, 2012, 2011, 2010, 2009  University of Chicago
##  created by Sam Quinan - samquinan@cs.uchicago.edu
##
##  Permission is hereby granted, free of charge, to any person obtaining
##  a copy of this software and associated documentation files (the
##  "Software"), to deal in the Software without restriction, including
##  without limitation the rights to use, copy, modify, merge, publish,
##  distribute, sublicense, and/or sell copies of the Software, and to
##  permit persons to whom the Software is furnished to do so, subject to
##  the following conditions:
##
##  The above copyright notice and this permission notice shall be
##  included in all copies or substantial portions of the Software.
##
##  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
##  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
##  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
##  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
##  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
##  OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
##  WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
##

import numpy, teem, ctypes

# translates from numpy type to teem type
def ndarrayGetTypes( array ):
    typeTable = {
        numpy.int8      :   (ctypes.c_byte, teem.nrrdTypeChar),
        numpy.uint8     :   (ctypes.c_ubyte, teem.nrrdTypeUChar),
        numpy.int16     :   (ctypes.c_short, teem.nrrdTypeShort),
        numpy.uint16    :   (ctypes.c_ushort, teem.nrrdTypeUShort),
        numpy.int       :   (ctypes.c_int, teem.nrrdTypeInt),
        numpy.uint      :   (ctypes.c_uint, teem.nrrdTypeUInt),
        numpy.int32     :   (ctypes.c_int32, teem.nrrdTypeInt),
        numpy.uint32    :   (ctypes.c_uint32, teem.nrrdTypeUInt),
        numpy.int64     :   (ctypes.c_int64, teem.nrrdTypeLLong),
        numpy.uint64    :   (ctypes.c_uint64, teem.nrrdTypeULLong),
        numpy.float     :   (ctypes.c_float, teem.nrrdTypeFloat),
        numpy.double    :   (ctypes.c_double, teem.nrrdTypeDouble)     }

    dt = array.dtype.type

    if dt not in typeTable:
        raise Exception("array type %s not supported" % dt)
    # sanity checking
    ct, tt = typeTable[dt]
    x = numpy.dtype(dt).itemsize
    y = ctypes.sizeof(ct)
    z = teem.nrrdTypeSize[tt]
    if not (x == y) or not (x == z) or not (y == z):
        raise Exception("corresponding numpy, teem, and ctypes types are not of the same size")
    return typeTable[dt]

def teemTypeToNumpyType(teem_type):
    numpy_endianess_glyph = '<' if teem.airMyEndian() == 1234 else '>'
    numpy_type_map = {
        teem.nrrdTypeChar: numpy_endianess_glyph + 'i1',
        teem.nrrdTypeUChar: numpy_endianess_glyph + 'u1',
        teem.nrrdTypeShort: numpy_endianess_glyph + 'i2',
        teem.nrrdTypeUShort: numpy_endianess_glyph + 'u2',
        teem.nrrdTypeInt: numpy_endianess_glyph + 'i4',
        teem.nrrdTypeUInt: numpy_endianess_glyph + 'u4',
        teem.nrrdTypeLLong: numpy_endianess_glyph + 'i8',
        teem.nrrdTypeULLong: numpy_endianess_glyph + 'u8',
        teem.nrrdTypeFloat: numpy_endianess_glyph + 'f4',
        teem.nrrdTypeDouble: numpy_endianess_glyph + 'f8'
    }
    return numpy_type_map[teem_type]


# design based on a combination of work done by Carlos Scheidegger
#   [ http://code.google.com/p/python-teem/source/browse/trunk/teem/capi/numpy/__init__.py ,
#     http://code.google.com/p/python-teem/source/browse/trunk/teem/nrrd.py ]
#   and ideas presented by Travis Oliphant [ http://blog.enthought.com/?p=62 ]

class Nrrd:
    def __init__(self):
        self._ctypesobj = teem.nrrdNew()
        # self._init = False
        self.base_ref = None
        self.teem = teem # world's ugliest hack -- but it appears to work

    def __del__(self):
        if self.base_ref == None:
            self.teem.nrrdNuke(self._ctypesobj)
        else:
            self.teem.nrrdNix(self._ctypesobj)
        self.base_ref = None
        self.teem = None

    def __get_array_interface(self):
        nrrd = self._ctypesobj.contents
        s = []
        for i in reversed(range(nrrd.dim)):
            s.append(nrrd.axis[i].size)
        r = {'shape': tuple(s),
            'typestr': teemTypeToNumpyType(nrrd.type),
            'data': (nrrd.data, False),
            'version': 3}
        return r

    __array_interface__ = property(__get_array_interface)

    def fromNDArray(self, array):

        if self.base_ref != None:
            self.base_ref = None
            teem.nrrdNix(self._ctypesobj)
            self._ctypesobj = teem.nrrdNew()

        try:
            cTy, teemTy = ndarrayGetTypes(array)
        except Exception as err:
            print err
            return

        c_type_p = ctypes.POINTER( cTy )
        if (array.flags.f_contiguous): # memory shared with array
            array_p = array.ctypes.data_as(c_type_p)
            dims = list(array.shape)
            sizes = (ctypes.c_size_t * teem.NRRD_DIM_MAX)(*dims)
            teem.nrrdWrap_nva(self._ctypesobj, array_p, teemTy, ctypes.c_uint(array.ndim), sizes)
            self.base_ref = array
        elif (array.flags.c_contiguous): # memory shared with array
            array_p = array.ctypes.data_as(c_type_p)
            dims = list(reversed(array.shape))
            sizes = (ctypes.c_size_t * teem.NRRD_DIM_MAX)(*dims)
            teem.nrrdWrap_nva(self._ctypesobj, array_p, teemTy, ctypes.c_uint(array.ndim), sizes)
            self.base_ref = array
        else: # must make c_contiguous -- memory not shared with array, transfer full control to teem
            arr = numpy.ascontiguousarray(array, dtype=array.dtype.type)
            array_p = arr.ctypes.data_as(c_type_p)
            dims = list(reversed(arr.shape))
            nrrd_tmp = teem.nrrdNew()
            sizes = (ctypes.c_size_t * teem.NRRD_DIM_MAX)(*dims)
            teem.nrrdWrap_nva(nrrd_tmp, array_p, teemTy, ctypes.c_uint(array.ndim), sizes)
            teem.nrrdCopy(self._ctypesobj, nrrd_tmp) # this call hangs with fmob-ch4.nrrd
            teem.nrrdNix(nrrd_tmp)
            self.base_ref = None
        self._init = True

        if isinstance(array, ExtendedArray):
            if not array.base_ref is None: # means there is some Nrrd file sharing it's memory with the array
                teem.nrrdAxisInfoCopy(self._ctypesobj, array.base_ref._ctypesobj, None, teem.NRRD_AXIS_INFO_SIZE_BIT)

    def load(self, file, args=None):

        if self.base_ref != None:
            self.base_ref = None
            teem.nrrdNix(self._ctypesobj)
            self._ctypesobj = teem.nrrdNew()

        teem.nrrdLoad(self._ctypesobj, file, args)
        self.base_ref = None
        self._init = True

    def save(self, file, args=None):
        teem.nrrdSave(file, self._ctypesobj, args)




class ExtendedArray(numpy.ndarray):
    def __new__(cls, input_array, base_ref=None):
        if isinstance(input_array, Nrrd):
            if input_array.base_ref == None:
                obj = numpy.asarray(input_array).view(cls)
                obj.base_ref = input_array
            else:
                obj = numpy.asarray(input_array.base_ref).view(cls)
                obj.base_ref = None
        else:
            obj = numpy.asarray(input_array).view(cls)
            obj.base_ref = base_ref
        return obj

    def __array_finalize__(self, obj):
        if obj is None: return
        self.base_ref = getattr(obj, 'base_ref', None)
